// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of scatter
//
// scatter is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// scatter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the
//       Free Software Foundation, Inc.
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <mathconst/mathconst.h>

#include <saotrace/io/fits/fits.h>
#include <randlib_c/randlib.h>

#include "EsaLinAsymDB.h"

namespace FITS = SAOTrace::IO::FITS;

namespace Scatter {

    bool
    EsaLinAsymDB::db_matches( const std::string& filename ) {

	FITS::TableFH table( filename );

	return table.has_column( "ESA" )
	    && table.has_column( "THETA" );
    }


    EsaLinAsymDB::EsaLinAsymDB( const std::string& filename) : ScatterDB( filename ) {

	std::vector <FITS::ColumnBase*> columns;

	FITS::Column<double> esa  ( "ESA"  , _esa,        columns );
	FITS::Column<double> theta( "THETA", _theta.data, columns );

	_nrows = FITS::read_table_columns( _file, columns );

	_theta.ncols = theta.extent[0];

    }

    bool
    EsaLinAsymDB::lookUpScatterAngle( double energy, double sin_alpha, double rannum, double& theta )
    {

	double esa = energy * sin_alpha;


	std::pair< ColData::iterator, ColData::iterator > range;
	std::pair< long, long > idx;

	bool is_extrapolated = find_bounds( _esa, esa, range, idx );


	if ( idx.first != idx.second )
	{
	    double theta_lo = lookUpTheta( idx.first, rannum );
	    double theta_hi = lookUpTheta( idx.second, rannum );
	    theta = interpolate( esa, *range.first, theta_lo, *range.second, theta_hi );
	}
	else
	{
	    theta = lookUpTheta( idx.first, rannum );
	}


	// convert theta to radians and flip its direction.  AlgBase::scatter
	// expects positive scattering angles in the backscatter direction.
	theta *= - M_PI  / ( 3600 * 180 );

	return is_extrapolated;
    }

    double
    EsaLinAsymDB::lookUpTheta( const long idx, double rannum ) {


	long ndist = _theta.ncols;

	double didx = rannum * (double) ndist;

	long jdx = (long) didx;

	if ( jdx >= ndist )
	    jdx = ndist - 1;

	double f2 = didx - (double) jdx;
	double f1 = 1 - f2;

	double theta = f1 * _theta( idx, jdx ) + f2 * _theta(idx, jdx + 1);

	return theta;
    }



}
