use strict;
use warnings;

use File::Basename;

use PDL;
use PDL::IO::FITS qw[ wfits ];

use Math::Trig qw( deg2rad pi );

use Astro::FITS::CFITSIO qw( :constants );
use Astro::FITS::CFITSIO::CheckStatus;

# manage the CFITSIO status variable
tie my $status, 'Astro::FITS::CFITSIO::CheckStatus';


my $base = basename( shift, '.fits');
my $fitsfile = "$base.fits";

unlink $fitsfile;
my $fptr = Astro::FITS::CFITSIO::create_file( $fitsfile, $status );

my $ntheta = 201;

my @cols = (
	    [ esa     => 'D'    ],
	    [ pdist   => $ntheta . 'D' ],
	    [ p0      => 'D'    ],
	    [ ptop    => 'D'    ],
	    [ pdelta  => 'D'    ],
	    [ powtop  => 'D'    ],
	    [ normtop => 'D'    ],
	   );

my %colpos = map { $cols[$_]->[0] => $_ + 1 } 0..@cols-1;

$fptr->create_tbl( BINARY_TBL, 0, scalar @cols,
		   [ map { $_->[0] } @cols ],
		   [ map { $_->[1] } @cols ],
		   undef,
		   'SCATTER',
		   $status );


$fptr->perlyunpacking(0);

#---------------------------
# grazing angle at which to run test.  The energy will be varied to vary
# the values of ESA
my $graze = 50;

my $esa_min = 1;
my $esa_max = 2;
my @esa = ( $esa_min, $esa_max );

#---------------------------
# write the ESA column
my $esa = pdl( @esa )->double;

my $expfmt = '%s_%d.exp.fits';

#---------------------------
# write out the theta distributions. at the same time write out
# expected theta distributions when the input ESA exactly matches
# the one in the table

my ( %prob, $theta );

my ( $rannum, $signs);
if ( @ARGV )
{
   ( $rannum, $signs ) =  rcols(shift);
}
else
{
    require PDL::GSL::RNG;
    $ENV{GSL_RNG_TYPE} = 'taus2';
    my $rng = PDL::GSL::RNG->new('default');
    $rannum = $rng->get_uniform(10000);

    $signs = $rannum->ones;
    $signs->mslice([0, $signs->nelem/2-1]) .= -1;
}

my $row = 1;

%prob = gen_prob( $ntheta, 0, 10 );
write_prob( $row, $esa, %prob );
my $theta1 = eval_theta( $rannum, %prob );
$theta1 *= $signs;
wfits( { theta => arcsec2rad($theta1) }, sprintf( $expfmt, $base, $row++) );

%prob = gen_prob( $ntheta, 0, 15 );
write_prob( $row, $esa, %prob );
my $theta2 = eval_theta( $rannum, %prob );
$theta2 *= $signs;
wfits( { theta => arcsec2rad( $theta2 ) }, sprintf( $expfmt, $base, $row++) );

$fptr->close_file($status);

#---------------------------
# now write out expected theta distributions for interpolated and
# extrapolated values of ESA

for my $esa ( 0.1, 1.5, 4 )
{
    push @esa, $esa;

    # don't extrapolate.  just use the distribution for esa_min
    # or esa_max
    if ( $esa <= $esa_min )
    {
	wfits( { theta => arcsec2rad($theta1->append(-$theta1)) } , sprintf( $expfmt, $base, $row++));
    }

    elsif ( $esa >= $esa_max )
    {
	wfits( { theta => arcsec2rad($theta2->append(-$theta2)) }, sprintf( $expfmt, $base, $row++));
    }

    else
    {
	my $theta = interpolate_theta( $esa_min, $esa_max,
				  $theta1, $theta2,  $esa );
	wfits( { theta => arcsec2rad($theta->append(-$theta)) }, sprintf( $expfmt, $base, $row++) );
    }
}

#---------------------------
# generate driver file for the test script.

open INPUTS, '>', $base . '.input'
  or die( "unable to open $base.input\n" );

$row = 1;
for my $esa ( @esa )
{
    my $energy = $esa / sin( deg2rad($graze / 60 ) );
    print INPUTS join("\t", $esa, $graze, $energy, 
		      sprintf( $expfmt, $base, $row++)), "\n";
}

close INPUTS;

sub arcsec2rad {

    return $_[0] / 3600 / 180  * pi;

}


sub write_prob
{
    my ( $row, $esa, %prob ) = @_;

    $prob{esa} = $esa->mslice([$row-1]);
    while( my ($col, $pos ) = each %colpos )
    {
	$fptr->write_col_dbl( $pos, $row, 1, $prob{$col}->nelem, 
			  $prob{$col}->get_dataref, $status );
    }

}


sub interpolate_theta
{
    my ( $x0, $x1, $y0, $y1, $x ) = @_;

    my $f1 = ($x - $x0) / ( $x1 - $x0 );
    my $f0 = 1 - $f1;

    return $f0 * $y0 + $f1 * $y1;
}


sub gen_prob
{
    my ( $ntheta, $theta_min, $theta_max ) = @_;

    # convert from arcseconds to radians
    $theta_min *= (pi / 180) / 3600;
    $theta_max *= (pi / 180) / 3600;

    my %prob;
    $prob{p0}     = pdl(0.0);
    $prob{ptop}   = pdl(0.9);
    $prob{pdelta} = ($prob{ptop} - $prob{p0}) / ($ntheta - 1);

    my $frac = (sequence($ntheta) +1 )/ $ntheta;
    $prob{pdist}  =  $theta_min + ($theta_max - $theta_min) * $frac * $frac;

    $prob{powtop} = pdl(-1/10);
    $prob{normtop} = $theta_max ** (1/$prob{powtop}) / ( 1 - $prob{ptop} );


    return %prob;
}


sub eval_theta
{
    my ( $x, %prob  ) =  @_;

    my $theta = $x->zeroes;

    my $tmp;
    my $p;

    $tmp = $theta->where( $x <= $prob{p0} );
    $tmp .= 0;

    ($tmp, $p) = where( $theta, $x, ( ($prob{p0} < $x) & ($x <= $prob{ptop}) ) );
    {
	my $jp = ( $p - $prob{p0} ) / $prob{pdelta};
	my $j = $jp->long;

	my $jt = $j->where( $j >= $prob{pdist}->nelem );
	$jt .= $prob{pdist}->nelem - 1;

	my $dj = $jp - $j;

	$tmp .= 
	      $prob{pdist}->index($j)   * ( 1 - $dj )
	    + $prob{pdist}->index($j+1) * $dj;
    }

    ($tmp, $p ) = where( $theta, $x, $x > $prob{ptop} );
    $p->inplace->clip( undef, 0.99999 );
    $tmp .= ($prob{normtop} * ( 1 - $p))**$prob{powtop};

    # convert theta to arcsec
    $theta *= 180 / pi * 3600;

    return ( $theta );
}


sub old_eval_theta
{
    my ( $ntheta, %prob  ) =  @_;

    my $theta = zeroes($ntheta);
    my $x     = random($ntheta);

    while( $ntheta-- )
    {
	my $p = $x->at($ntheta);

	if ( $p <= $prob{p0} )
	{
	    $theta->set($ntheta, 0);
	}
	elsif ( $prob{p0} < $p && $p <= $prob{ptop} )
	{
	    my $jp = ( $p - $prob{p0} ) / $prob{pdelta};
	    my $j = int($jp);

	    $j = $prob{pdist}->nelem - 1
	      if $j >= $prob{pdist}->nelem;

	    my $dj = $jp - $j;

	    $theta->set($ntheta,
			$prob{pdist}->at($j)   * ( 1 - $dj)
			+ $prob{pdist}->at($j+1) * $dj
		       );
	}
	else
	{
	    $p = 0.99999 if $p > 0.99999;
	    $theta->set($ntheta, ($prob{normtop} * ( 1 - $p))**$prob{powtop} );
	}

    }
}
