use strict;
use warnings;

use File::Basename;

use PDL;
use PDL::IO::FITS qw[ wfits ];

use Math::Trig qw( deg2rad pi );

use Astro::FITS::CFITSIO qw( :constants );
use Astro::FITS::CFITSIO::CheckStatus;

# manage the CFITSIO status variable
tie my $status, 'Astro::FITS::CFITSIO::CheckStatus';


my $base = basename( shift, '.fits');
my $fitsfile = "$base.fits";

unlink $fitsfile;
my $fptr = Astro::FITS::CFITSIO::create_file( $fitsfile, $status );

$fptr->create_tbl( BINARY_TBL, 0, 2, 
		   [ 'ESA', 'THETA' ],
		   [ 'D', '1001D' ],
		   undef,
		   'SCATTER',
		   $status );


$fptr->perlyunpacking(0);

#---------------------------
# grazing angle at which to run test.  The energy will be varied to vary
# the values of ESA
my $graze = 50;
my $sin_graze = sin( deg2rad($graze / 60 ) );
my $esa_min = 1;
my $esa_max = 2;
my @esa = ( $esa_min, $esa_max );
my $idx = 1;

my $n_op = 10000;

#---------------------------
# write the ESA column
my $esa = pdl( @esa )->double;
$fptr->write_col_dbl( 1, 1, 1, 2, $esa->get_dataref, $status );

my $expfmt = "${base}_%d.exp";

#---------------------------
# write out the theta distributions. at the same time write out
# expected theta distributions when the input ESA exactly matches
# the one in the table

my @theta1_p = ( -10, 10, 5 );

my $theta1_i   = in_plane ( @theta1_p );
my $theta1_i_p = in_plane ( @theta1_p, $n_op );
my $theta1_o_p = out_of_plane( @theta1_p, $n_op );

$fptr->write_col_dbl( 2, 1, 1, $theta1_i->nelem, $theta1_i->get_dataref, $status );

write_probs( $theta1_i_p, $theta1_o_p, $expfmt, $idx++ );

my @theta2_p = ( -10, 10, -5 );

my $theta2_i   = in_plane( @theta2_p );
my $theta2_i_p = in_plane( @theta2_p, $n_op );
my $theta2_o_p = out_of_plane( @theta2_p, $n_op );

$fptr->write_col_dbl( 2, 2, 1, $theta2_i->nelem, $theta2_i->get_dataref, $status );

write_probs( $theta2_i_p, $theta2_o_p, $expfmt, $idx++ );

$fptr->close_file($status);


#---------------------------
# now write out expected theta distributions for interpolated and
# extrapolated values of ESA

for my $esa ( 0.1, 1.5, 4 )
{
    push @esa, $esa;

    # don't extrapolate.  just use the distribution for esa_min
    # or esa_max
    if ( $esa <= $esa_min )
    {
	write_probs( $theta1_i_p, $theta1_o_p, $expfmt, $idx++ );
    }

    elsif ( $esa >= $esa_max )
    {
	write_probs( $theta2_i_p, $theta2_o_p, $expfmt, $idx++ );
    }

    else
    {
	my $theta = interpolate_theta( $esa_min, $esa_max, $theta1_i_p, $theta2_i_p,  $esa );

	my $theta_x = interpolate_theta( $esa_min, $esa_max, $theta1_i_p, $theta2_i_p,  $esa );

	write_probs( $theta, out_of_plane( $theta_x ), $expfmt, $idx++ );
    }
}


#---------------------------
# generate driver file for the test script.

open INPUTS, '>', $base . '.input'
  or die( "unable to open $base.input\n" );

$idx = 1;
for my $esa ( @esa )
{
    my $energy = $esa / $sin_graze;
    print INPUTS join("\t", $esa, $graze, $energy, 
		      sprintf( $expfmt, $idx) . '.fits'
		     ), "\n";
}
continue{
    $idx++;
}

close INPUTS;


sub interpolate_theta
{
    my ( $x0, $x1, $y0, $y1, $x ) = @_;

    my $f1 = ($x - $x0) / ( $x1 - $x0 );
    my $f0 = 1 - $f1;

    return $f0 * $y0 + $f1 * $y1;
}


sub in_plane
{
    my ( $theta_min, $theta_max, $center, $nprob ) = @_;

    $nprob ||= 1001;

    my $theta = pdl( $theta_min, $center, $theta_max );
    my $pdf = pdl( 0, 100, 0);
    my $thetas = ($theta_max - $theta_min) * sequence( $nprob ) / ($nprob - 1 ) + $theta_min;
    $pdf = interpol( $thetas, $theta, $pdf );

    my $cdf = $pdf->cumusumover;
    $cdf /= $cdf->max;

    my $prob = sequence($nprob) / ($nprob - 1);

    my $theta_i = interpol( $prob, $cdf, $thetas );

    return $theta_i;
}

sub out_of_plane {

    my ( $theta_ip ) = @_ > 1 ? in_plane( @_ ) : $_[0];

    # each theta has the same probability.  we are adding more thetas such that 
    # -theta has the same probability as +theta.
    my $theta = $theta_ip->append( -$theta_ip );

    # Some of the theta's might be duplicates; need to resolve those by
    # constructing a new CDF, removing the duplicates, then generating a new
    # set of theta sampled from the CDF.

    # sort array
    $theta = $theta->qsort;

    # and the CDF
    my $cdf = $theta->ones->cumusumover;
    $cdf /= $cdf->max;

    # If there's a run of duplicates theta_0...theta_n, the real
    # CDF is $cdf->at(theta_n), as it'll be the sum of the pdf's
    # for the similar theta's.

    # this'll return the number of times each theta is present
    my ( $di, $utheta ) = $theta->rle;

    # strip off unneeded slots
    ( $di, $theta ) = map { $_->sever } where( $di, $utheta, $di > 0 );

    # turn di into an index which points at the LAST duplicate
    # theta, which will be the true CDF for theta
    $di = $di->cumusumover - 1;

    # and reconstruct $cdf and $thetas;
    $cdf = $cdf->index($di)->sever;

    my $prob = $theta_ip->sequence / ($theta_ip->nelem - 1);

    return $sin_graze * interpol( $prob, $cdf, $theta );
}

sub arcsec2rad {

    return $_[0] / 3600 / 180  * pi;

}

sub write_probs {

    my ( $in_plane, $out_of_plane, $fmt, $idx ) = @_;

    wfits( { theta => arcsec2rad( $in_plane),
	     phi => arcsec2rad( $out_of_plane)
	   }, sprintf( $fmt, $idx) . '.fits' );
}
