#! /usr/bin/env perl
# -*-perl-*-

use strict;
use warnings;

use Getopt::Long;
use Math::Trig qw[ pi deg2rad ];

use Astro::FITS::CFITSIO::Simple qw[ rdfits ];
use PDL qw[ norm inner crossp ];
use PDL::IO::FITS qw[ wfits];

use IO::File;

use RDBLite;

our %opt;

our $dbroot = "$ENV{INPUTDIR}/databases";

eval { main() };

do { warn $@; exit 1; } if $@;

exit 0;

sub main
{
  parse_args();

  scatter_rays();
  my $results = calc_angles( $opt{tag} . '.flat.fits', $opt{tag} . '.scatter.fits' );

  wfits $results , $opt{tag} . ".scatter.calc.fits";

  my %stats;
  @stats{ qw[ mean prms median min max adev rms ] } = $results->{dtheta}->stats;

  hdump( $opt{tag}. ".dtheta", %stats );

  @stats{ qw[ mean prms median min max adev rms ] } = $results->{dphi}->stats;

  hdump( $opt{tag}. ".dphi", %stats );

}

sub hdump {

    my ( $file, %hash ) = @_;

    my $fh = IO::File->new( $file, 'w' )
      or die( "unable to create $file\n" );

    $fh->printf( "%s: %s\n", $_, $hash{$_} ) foreach keys %hash;
}

sub scatter_rays
{
    my $trans_file = "$ENV{INPUTDIR}/events.rdb";

  # generate photons.  theta = - alpha because we're generating the _reflected ray_
  runi( $ENV{GENPHOT},
	spectrum	=> 'none',
	output		=> $opt{tag} . ".genphot.bp",
	geo		=> 'point',
	dist		=> 'uniform',
	energy_dist	=> 'min',
	photdens	=> $opt{nrays},
	seed1		=> $opt{seed1},
	seed2		=> $opt{seed1},
	block		=> $opt{block},
	x		=> 0,
	y		=> 0,
	z		=> $opt{zpos},
	phi		=> 0,
	ray_theta	=> -$opt{alpha},
	ray_phi		=> 0,
	height		=> 100,
	width		=> 100,
	energy		=> $opt{energy},
	version		=> 'no',
	help		=> 'no',
      );


  # intersect them with the flat.  all this does is add the surface normal vector to the
  # bpipe stream.
  runi( $ENV{FLAT},
	input		=> $opt{tag} . '.genphot.bp',
	output		=> $opt{tag} . '.flat.bp',
	x0		=> 0,
	y0		=> 0,
	z0		=> $opt{zpos},
	normx		=> 1,
	normy		=> 0,
	normz		=> 0,
	height		=> 100,
	width		=> 100,
	rotation	=> 0,
	surf_no		=> $opt{surfno},
	help		=> 'no',
	version		=> 'no',
      );

  # convert to FITS
  runi( $ENV{BP2FITS},
	input	=> $opt{tag} . '.flat.bp',
	output	=> $opt{tag} . '.flat.fits',
	trans   => $trans_file,
	add_exact_dpktfs  => 'id,direction,surface_norm',
	vectors => 'yes',
      );


  # scatter 'em
  runi( $ENV{SCATTER},
	input		=> $opt{tag} . '.flat.bp',
	output		=> $opt{tag} . '.scatter.bp',
	logfile		=> $opt{tag} . '.scatter.log',
	seed1		=> $opt{seed1},
	seed2		=> $opt{seed1},
	block		=> ++$opt{block},
	scatter_db	=> $opt{scatter_db},
	mirror_db	=> $opt{mirror_db},
	scat_iter_lim	=> 10,
	in_plane	=> $opt{in_plane} ? 'yes' : 'no',
	out_of_plane	=> $opt{out_of_plane} ? 'yes' : 'no',
	min_cumulative_probability		=> 0.0,
	max_cumulative_probability		=> 1.0,
	shell		=> $opt{shell},
	surf_no		=> $opt{surfno},
	theta		=> $opt{theta},
	phi		=> $opt{phi},
	debug		=> 'theta,phi',
	help		=> 'no',
	usage		=> 'no',
	version		=> 'no',
      );

  # convert to FITS
  runi( $ENV{BP2FITS},
	input	=> $opt{tag} . '.scatter.bp',
	output	=> $opt{tag} . '.scatter.fits',
	trans   => $trans_file,
	add_exact_dpktfs  => 'id,direction,surface_norm,theta,phi',
	vectors => 'yes',
      );
}



# run programs with IRAF style parameters
sub runi
{
    my ( $command, %args ) = @_;

    my @args;
    while ( my ( $option, $value ) = each %args )
    {
	push @args, $option . '=' . $value ;
    }

    print STDERR join(' ', $command, @args ), "\n"
      if $opt{verbose};

    system( $command, @args ) == 0
      or die( "error running ", join( ' ', $command, @args ), "\n" );
}

sub run
{
    my ( $command, @args ) = @_;

    print STDERR join(' ', $command, @args ), "\n"
      if $opt{verbose};

    system( $command, @args ) == 0
      or die( "error running ", join( ' ', $command, @args ), "\n" );
}

sub read_ray_angles {

    my ( $flat_fits, $scatter_fits ) = @_;

    my %flat = rdfits( $flat_fits );
    $_ = $_->squeeze foreach values %flat;

    my %scat = rdfits( $scatter_fits );

    # PDL 2.7 has problems with PDL->null->squeeze, generating
    # piddles which return an error on further operations.
    # so query the number of elements before we squeeze.
    return {}, {}
      if $scat{id}->nelem == 0;

    $_ = $_->squeeze foreach values %scat;


    # it's possible that a ray got axed, so create
    # proper indices to line up flat_db and scat_db values
    if ( $flat{id}->nelem != $scat{id}->nelem ||
         any ( $flat{id} != $scat{id} ) )
    {
	# ensure that $flat{id} is sequential. can't imagine
	# why it shouldn't be!
	my $diff = $flat{id} - $flat{id}->rotate(1);
	die( "non-sequential flat ids\n" )
	  if $flat{id}->at(0) != 1 ||
	    $diff->at(0) != - ($flat{id}->nelem - 1);

	$diff->set(0,1);

	die( "non-sequential flat ids\n" )
	  if $diff->at(0) != 1;

	# regenerate flat piddles using the scat id as an index
	my $idx = $scat{id} - 1;

	$flat{$_} = $flat{$_}->index($idx)->sever foreach keys %flat;
    }

    return \%flat, \%scat;
}


# calculate the scattering angles (in and out of plane)
# this calculation is closely tied to the definition of the angles,
# so if (when) the definitions in the scatter code (MZ.cc: scatter) change,
# this code must also change.
sub calc_angles
{
    my ( $flat, $scat ) = read_ray_angles( @_ );

    my $theta  = null();
    my $dtheta = null();
    my $phi    = null();
    my $dphi   = null();

    if ( keys %$scat )
    {

	my $u = norm(crossp( $flat->{dir}, $scat->{norm} ));
	my $v = crossp( $scat->{norm}, $u );

	my $l = ($v * inner($scat->{dir}, $v)->dummy(0,3) + $u * inner($scat->{dir}, $u)->dummy(0,3) )->norm;

	$theta = asin ( - inner( $scat->{dir}, $scat->{norm} ) ) - deg2rad($opt{alpha});
	$phi   = acos( inner( $l, $v ) ) * ( -2 * (inner( $l, $u) < 0 ) + 1 );

	$dtheta = $theta - $scat->{theta};
	$dphi   = $phi - $scat->{phi};

	if ( $opt{flip} )
	{
	    $theta *= -1;
	    $phi *= -1;
	}


    }

    return { theta  => $theta,
	     dtheta => $dtheta,
	     phi    => $phi,
	     dphi   => $dphi,
	   };
}

sub parse_args
{

    %opt =
      (
       alpha      => 51.259,
       block	  => 0,
       energy     => 4.51,
       flip	  => 0,
       in_plane	  => 0,
       mirror_db  => undef,
       nrays	  => 1000,
       out_of_plane => 0,
       phi	  => 0,
       scatter_db => undef,
       seed1	  => 1,
       seed2	  => 1,
       shell      => 1,
       surfno     => 1,
       tag	  => 'foo',
       theta	  => 0,
       verbose    => 0,
       zpos       => -432.9811,
      );

    eval {
	local $SIG{ __WARN__ } = sub { die $_[0] };

	GetOptions ( \%opt, 
		     qw/
			alpha=f
			block=i
			energy=f
			flip
			in_plane!
			mirror_db=s
			nrays=i
			out_of_plane!
			phi=f
			scatter_db=s
			seed1=i
			seed2=i
			shell=i
			srcdir=s
			surfno=i
			tag=s
			theta=f
			verbose
			zpos=f
			/,
		   );
    };
    die $@ if $@;

    # alpha: arcmin => degrees
    $opt{alpha} /= 60;

    $opt{geometry} = (qw( p h ))[$opt{surfno}-1];

    die( "illegal surface number: $opt{surfno} \n" )
      unless defined $opt{geometry};

    $opt{optic} = $opt{geometry} . $opt{shell};

    # energy * sine(alpha)
    $opt{energy_sine_alpha} = $opt{energy} * sin( deg2rad($opt{alpha}) );

    my $mirror = 
      RDB::Lite::match_record( str_interpolate($opt{mirror_db}),
			       mirror => $opt{optic} )
	  or  die( "can't find optic ($opt{optic}) in $opt{mirror_db}\n" );

    $opt{scatter_tbl} = get_scatter_table( str_interpolate($opt{scatter_db}),
					   $mirror->{z0} )
      or die( "error extracting scatter table name from $opt{scatter_db}\n" );

    $ENV{srcdir} = $opt{srcdir} if defined $opt{srcdir};


    $opt{scatter_tbl} = str_interpolate($opt{scatter_tbl});

    my @notset = grep { !defined $opt{$_} } keys %opt;
    die( 'parameters `', join( "`, `", @notset ), "' are not set\n" )
      if @notset;

}

sub get_scatter_table
{
    my ( $db, $z0 ) = @_;


    my $rdb = RDB::Lite->new( $db );
    while( my $data = $rdb->read )
    {
	return $data->{scat_file}
	  if $data->{mirror} eq $opt{optic} &&
	     $data->{zmin}   <  $opt{zpos} - $z0 &&
	     $data->{zmax}   >  $opt{zpos} - $z0
    }

    return;
}



#----------------------------------------------------------------#

sub str_interpolate
{
  my ( $text, $var ) = @_;

  $var ||= \%ENV;

  $text =~ s{
       \$                         # find a literal dollar sign
      (?:{(\w+)}|(\w+))           # find a "word" and store it in $1 or $2
  }{
      my $t = defined $2 ? $2 : $1;
      defined $var->{$t} ? $var->{$t} : '';
  }egx;

  return $text;
}


